use std::ffi::CString;

use bitflags::bitflags;
use implot_sys as sys;
use sys::{ImAxis, ImVec2, *};

use crate::{Context, PlotUi};

const DEFAULT_PLOT_SIZE_X: f32 = 400.0;
const DEFAULT_PLOT_SIZE_Y: f32 = 400.0;

pub enum Axis {
	X1 = 0,
	X2 = 1,
	X3 = 2,
	Y1 = 3,
	Y2 = 4,
	Y3 = 5,
}

bitflags! {
	#[repr(transparent)]
	pub struct PlotFlags: u32 {
		const NONE = ImPlotFlags__ImPlotFlags_None;
		const NOTITLE = ImPlotFlags__ImPlotFlags_NoTitle;
		const NOLEGEND = ImPlotFlags__ImPlotFlags_NoLegend;
		const NOMOUSETEXT = ImPlotFlags__ImPlotFlags_NoMouseText;
		const NOINPUTS = ImPlotFlags__ImPlotFlags_NoInputs;
		const NOMENUS = ImPlotFlags__ImPlotFlags_NoMenus;
		const NOBOXSELECT = ImPlotFlags__ImPlotFlags_NoBoxSelect;
		const NOCHILD = ImPlotFlags__ImPlotFlags_NoChild;
		const NOFRAME = ImPlotFlags__ImPlotFlags_NoFrame;
		const EQUAL = ImPlotFlags__ImPlotFlags_Equal;
		const CROSSHAIRS = ImPlotFlags__ImPlotFlags_Crosshairs;
		const CANVASONLY = ImPlotFlags__ImPlotFlags_CanvasOnly;
	}
}

bitflags! {
	#[repr(transparent)]
	pub struct PlotAxisFlags: u32 {
		const NONE = ImPlotAxisFlags__ImPlotAxisFlags_None;
		const NO_LABEL = ImPlotAxisFlags__ImPlotAxisFlags_NoLabel;
		const NO_GRID_LINES = ImPlotAxisFlags__ImPlotAxisFlags_NoGridLines;
		const NO_TICK_MARKS = ImPlotAxisFlags__ImPlotAxisFlags_NoTickMarks;
		const NO_TICK_LABELS = ImPlotAxisFlags__ImPlotAxisFlags_NoTickLabels;
		const NO_INITIAL_FIT = ImPlotAxisFlags__ImPlotAxisFlags_NoInitialFit;
		const NO_MENUS = ImPlotAxisFlags__ImPlotAxisFlags_NoMenus;
		const NO_SIDESWITCH = ImPlotAxisFlags__ImPlotAxisFlags_NoSideSwitch;
		const NO_HIGHLIGHT = ImPlotAxisFlags__ImPlotAxisFlags_NoHighlight;
		const OPPOSITE = ImPlotAxisFlags__ImPlotAxisFlags_Opposite;
		const FOREGROUND = ImPlotAxisFlags__ImPlotAxisFlags_Foreground;
		const INVERT = ImPlotAxisFlags__ImPlotAxisFlags_Invert;
		const AUTO_FIT = ImPlotAxisFlags__ImPlotAxisFlags_AutoFit;
		const RANGE_FIT = ImPlotAxisFlags__ImPlotAxisFlags_RangeFit;
		const PAN_STRETCH = ImPlotAxisFlags__ImPlotAxisFlags_PanStretch;
		const LOCK_MIN = ImPlotAxisFlags__ImPlotAxisFlags_LockMin;
		const LOCK_MAX = ImPlotAxisFlags__ImPlotAxisFlags_LockMax;
		const LOCK = ImPlotAxisFlags__ImPlotAxisFlags_Lock;
		const NO_DECORATIONS = ImPlotAxisFlags__ImPlotAxisFlags_NoDecorations;
		const AUX_DEFAULT = ImPlotAxisFlags__ImPlotAxisFlags_AuxDefault;
	}
}

pub struct Plot {
	title: CString,
	size: [f32; 2],
	x_label: CString,
	y_label: CString,
	flags: ImPlotFlags,
	x_flags: ImPlotAxisFlags,
	y_flags: ImPlotAxisFlags,
}

impl Plot {
	pub fn new(title: &str) -> Self {
		Self {
			title: CString::new(title)
				.unwrap_or_else(|_| panic!("string contains internal null bytes: {}", title)),
			size: [DEFAULT_PLOT_SIZE_X, DEFAULT_PLOT_SIZE_Y],
			x_label: CString::new("").unwrap(),
			y_label: CString::new("").unwrap(),
			flags: PlotFlags::NONE.bits() as ImPlotFlags,
			x_flags: PlotAxisFlags::NONE.bits() as ImPlotAxisFlags,
			y_flags: PlotAxisFlags::NONE.bits() as ImPlotAxisFlags,
		}
	}

	pub fn size(mut self, size: [f32; 2]) -> Self {
		self.size = size;
		self
	}

	pub fn with_flags(mut self, flags: &PlotFlags) -> Self {
		self.flags = flags.bits() as ImPlotFlags;
		self
	}

	pub fn with_x_flags(mut self, flags: &PlotAxisFlags) -> Self {
		self.x_flags = flags.bits() as ImPlotAxisFlags;
		self
	}

	pub fn with_y_flags(mut self, flags: &PlotAxisFlags) -> Self {
		self.y_flags = flags.bits() as ImPlotAxisFlags;
		self
	}

	pub fn x_label(mut self, label: &str) -> Self {
		self.x_label = CString::new(label)
			.unwrap_or_else(|_| panic!("string contains internal null bytes: {}", label));

		self
	}

	pub fn y_label(mut self, label: &str) -> Self {
		self.y_label = CString::new(label)
			.unwrap_or_else(|_| panic!("string contains internal null bytes: {}", label));

		self
	}

	pub fn begin(&self, plot_ui: &PlotUi) -> Option<PlotToken> {
		let should_render = unsafe {
			let size_vec: ImVec2 = ImVec2 {
				x: self.size[0],
				y: self.size[1],
			};

			let should_render = sys::ImPlot_BeginPlot(self.title.as_ptr(), size_vec, self.flags);

			should_render
		};

		if should_render {
			unsafe {
				sys::ImPlot_SetupAxis(Axis::X1 as i32, self.x_label.as_ptr(), self.x_flags);
				sys::ImPlot_SetupAxis(Axis::Y1 as i32, self.y_label.as_ptr(), self.y_flags);
			}

			Some(PlotToken {
				context: plot_ui.context,
				plot_title: self.title.clone(),
			})
		} else {
			None
		}
	}

	pub fn build<F: FnOnce()>(self, plot_ui: &PlotUi, f: F) {
		if let Some(token) = self.begin(plot_ui) {
			f();
			token.end()
		}
	}
}

pub struct PlotToken {
	context: *const Context,
	plot_title: CString,
}

impl PlotToken {
	pub fn end(mut self) {
		self.context = std::ptr::null();
		unsafe { sys::ImPlot_EndPlot() };
	}
}

impl Drop for PlotToken {
	fn drop(&mut self) {
		if !self.context.is_null() && !std::thread::panicking() {
			panic!(
				"warning a plottoken for plot \"{:?}\" was not called end() on.",
				self.plot_title
			);
		}
	}
}
